from .scene import LidarScene
from .transform import Transform
from .frame import LidarFrame
from .camera import LidarCamera
from .image import LidarImage
from .helper import (
    s3_smart_upload,
    format_lidar_point,
    format_point,
    format_quaternion,
    scale_file_upload,
    get_signed_url
)

from io import BytesIO
from functools import partial
from multiprocessing.pool import ThreadPool
from typing import MutableMapping, List, Dict
from urllib.parse import urlparse
from typing import Any

import pandas as pd
import numpy as np

import ujson
import nucleus

UPLOAD_POOL_SIZE = 8

#Formatting Helpers

def get_image_url_path(path, camera_id, frame_id):
    return f"{path}/image-{camera_id}-{frame_id}.jpg"

def get_image_ref_id(ref_id_prefix, camera_id, frame_id):
    return f"{ref_id_prefix}-{camera_id}-image-{frame_id}"

def get_pointcloud_ref_id(ref_id_prefix, frame_id):
    return f"{ref_id_prefix}-pointcloud-{frame_id}"

def get_scene_ref_id(ref_id_prefix):
    return f"{ref_id_prefix}-scene"

class Cuboid():
    '''Work in progress class - need to fill out'''
    def __init__(self, position: np.ndarray, yaw: float):
        self.position: np.ndarray = np.zeroes((0,3), dtype = float)
        self.yaw: float = None

class NucleusLidarFrame(LidarFrame):
    '''Overloaded Nucleus Frame class

    Annotation and predictions are next up for implementation
    '''

    def __init__(self, frame_id, cameras):
        self.id = frame_id
        self.cameras: pd.Series[Any, LidarCamera] = cameras
        self.images: pd.Series[Any, LidarImage] = pd.Series(dtype=object)
        self.points: np.ndarray = np.zeros((0, 5), dtype=float)
        self.pointcloud_metadata: dict = {}
        self.radar_points: np.ndarray = np.zeros((0, 3), dtype=float)
        self.colors: np.ndarray = np.zeros((0, 3), dtype=float)
        self.transform = Transform()
        self.annotations: pd.Series[Any,Cuboid] = pd.Series(dtype=object)
        self.predictions: pd.Series[Any,Cuboid] = pd.Series(dtype=object)

    def add_points(
        self, points: np.array, transform: Transform = None, metadata: dict = None, intensity=1, sensor_id=0,
    ):
        """Add points to the frame, structure: np.array with dimension 1 and shape (N,3) or (N,4) (N being the number of point in the frame)

        Points with intensity:

        .. highlight:: python
        .. code-block:: python

          points = np.array([
            [0.30694541, 0.27853175, 0.51152715, 0.4],
            [0.80424087, 0.24164057, 0.45256181, 1],
            ...
          ])

        Points without intensity:

        .. highlight:: python
        .. code-block:: python

          points = np.array([
            [0.30694541, 0.27853175, 0.51152715],
            [0.80424087, 0.24164057, 0.45256181],
            ...
          ])

        :param points: List of points
        :type points: np.array
        :param transform: Transform that should be applied to the points
        :type transform: Transform
        :param metadata: Any pointcloud metadata to be associated with dataset item
        :type metadata: dict
        :param intensity: If the points list doesn't include intensity, this value will be used as intensity for all the points (default ``1``)
        :type intensity: int
        :param sensor_id: Sensor id, used in case that you have more than one lidar sensor. (Default ``0``)
        :type sensor_id: int

        """
        if points.ndim == 1:
            points = np.array([points])
        if points.shape[1] == 3:
            points = np.hstack([points, np.ones((points.shape[0], 1)) * intensity])
        if transform is not None:
            points = transform.apply(points)

        points = np.hstack([points, np.ones((points.shape[0], 1)) * sensor_id])

        self.points = np.vstack([self.points, points])
        self.pointcloud_metadata = metadata

    def to_json(
        self,
        base_url: str = "",
        s3_upload: bool = True,
        project_name: str = ""
    ):
        """Returns pointcloud json

        :param base_url: This url will concatenated with the frame name
        :type base_url: str
        :returns: Frame object as a JSON formatted stream
        :rtype: str

        """

        points_json = pd.DataFrame(
            self.get_world_points(), columns=["x", "y", "z", "i", "d"]
        ).to_json(double_precision=4, orient="records", date_format=None)

        frame_object = {
            "points": "__POINTS__",
            "device_position": format_point(self.transform.position),
            "device_heading": format_quaternion(self.transform.quaternion),
        }

        out = ujson.dumps(frame_object)
        out = out.replace('"__POINTS__"', points_json)

        return out

    def s3_upload(self, bucket: str, path: str):
        """Save frame in S3

        :param bucket: S3 Bucket name
        :type bucket: str
        :param path: Path to store data
        :type key: str
        """
        # print(f'Uploading frame {self.id}...')
        base_url = f"s3://{bucket}/{path}"

        # Upload frame json file
        s3_smart_upload(
            fileobj=BytesIO(
                bytes(self.to_json(base_url), encoding="utf-8")
            ),
            bucket=bucket,
            key=f"{path}/frame-{self.id}.json",
            content_type="application/json",
        )

        # Upload images
        for camera_id, image in self.images.items():
            image.s3_upload(bucket, f"{path}/image-{camera_id}-{self.id}.jpg")

    def generate_cam_nucleus_dataset_items(
        self,
        scene_dict: dict,
        ref_id_prefix: str,
        presigned_items: bool = False
    ):
        """Generates all necessary camera dataset items for corresponding LidarFrame

        :param scene_dict: Mapping from frame and camera images to URL
        :type scene_dict: str
        :param ref_id_prefix: String insert at beginning of automatically generated ref-id, required
        :type ref_id_prefix: str
        :param presigned_items: Presigns all URLs via S3
        :type presigned_items: str
        :returns: Dictionary of Nucleus image dataset items associated with camera ID
        :rtype: Dict

        """
        assert ref_id_prefix is not "", "Please set a Reference ID prefix to ensure reference idempotency."

        def generate_camera_params(self, camera):
            """Generates camera specific metadata for nucleus dataset item"""
            wct = self.transform @ camera.pose

            heading = format_quaternion(wct.quaternion)
            position = format_point(wct.translation)
            camParams = {
                "cx": float(camera.cx),
                "cy": float(camera.cy),
                "fx": float(camera.fx),
                "fy": float(camera.fy),
                "k1": float(camera.D[0]),
                "k2": float(camera.D[1]),
                "p1": float(camera.D[2]),
                "p2": float(camera.D[3]),
                "k3": float(camera.D[4]),
                "k4": float(camera.D[5]),
                "k5": float(camera.D[6]) if len(camera.D) >= 7 else 0,
                "k6": float(camera.D[7]) if len(camera.D) >= 8 else 0,
                "heading": heading,
                "position": position,
                "camera_model": camera.model,
            }
            return camParams

        nucleus_camera_items = {}
        for camera in self.cameras:
            camera_params = generate_camera_params(self, camera)
            image_location = scene_dict["cameras"][camera.id][self.id]

            if presigned_items:
                image_location = get_signed_url(
                    bucket=urlparse(image_location).netloc,
                    path=urlparse(image_location).path[1:],
                )

            item_metadata = {"camera_params": camera_params}

            if self.images[camera.id].metadata:
                item_metadata = dict(self.images[camera.id].metadata, **item_metadata)

            item = nucleus.DatasetItem(
                image_location=image_location,
                reference_id=get_image_ref_id(ref_id_prefix,camera.id,self.id),
                metadata=item_metadata,
            )
            nucleus_camera_items[str(camera.id)] = item

        return nucleus_camera_items


class NucleusLidarScene(LidarScene):
    '''Overloaded Nucleus scene'''
    def __init__(self):
        """
        :rtype: object
        """
        self.cameras: MutableMapping[LidarCamera] = pd.Series()
        self.frames: MutableMapping[NucleusLidarFrame] = pd.Series()
        self.base_url = None
        self.scale_file_attachments = None
        self.ref_id_prefix = ""

    def from_LidarScene(self, LidarScene):
        self.cameras = LidarScene.cameras
        for frame_idx, LidarFrame in enumerate(LidarScene.frames):
            self.get_frame(frame_idx).points = LidarFrame.points
            self.get_frame(frame_idx).images = LidarFrame.images
            self.get_frame(frame_idx).transform = LidarFrame.transform
        return self

    def set_ref_id_prefix(self, ref_id_prefix: str):
        self.ref_id_prefix = ref_id_prefix

    def to_dict(self, base_url: str = None) -> dict:
        """Return a dictionary with the frame urls using the base_url as base.

        :param base_url: This url will be concatenated with the frames name, e.g.: `'%s/frame-%s.json' % (base_url, frame.id)`
        :type base_url: str
        :return: Dictionary with the frame urls data
        :rtype: dict

        """
        if base_url is None:
            base_url = self.base_url

        cameras = {}
        for camera in self.cameras:
            cameras[camera.id] = [
                "%s/image-%s-%s.jpg" % (base_url, camera.id, frame.id)
                for frame in self.frames
            ]

        return dict(
            frames=["%s/frame-%s.json" % (base_url, frame.id) for frame in self.frames],
            cameras=cameras,
        )

    def get_frame(self, frame_id=None, index: int = None) -> NucleusLidarFrame:
        """Get a frame by id (or index) or create one if it does not exist

        :param frame_id: The frame id
        :type frame_id: str, int
        :param index: The frame index
        :type index: int

        :return: NucleusLidarFrame
        :rtype: NucleusLidarFrame

        """
        assert (
            frame_id is not None or index is not None
        ), "id or index must be specified"

        if frame_id is None:
            frame_id = self.frames.index[index]

        if frame_id not in self.frames:
            if isinstance(frame_id, int):
                self.frames.index = self.frames.index.astype(int)
            self.frames[frame_id] = NucleusLidarFrame(frame_id, cameras=self.cameras)
        return self.frames[frame_id]

    def s3_upload(
        self,
        bucket: str,
        path=None,
        mock_upload: float = False,
        use_threads: float = True
    ):
        """Overloaded S3 upload function

        :param bucket: S3 Bucket name
        :type bucket: str
        :param path: Path to store data
        :type key: str
        :param mock_upload: To avoid upload the data to S3  (defualt ``False``)
        :type mock_upload: float
        :param use_threads: In order to upload multiple files at the same time using threads  (defualt ``True``)
        :type use_threads: float

        :return: Scene S3 url
        :rtype: str

        """
        self.base_url = f"s3://{bucket}/{path}"

        print("Uploading scene to S3: %s" % self.base_url)
        scene_dict = self.to_dict(self.base_url)

        poses_csv = pd.DataFrame(
            self.frames.map(lambda f: list(f.transform.matrix.reshape(-1))).to_dict()
        ).T.to_csv(header=False)

        if not mock_upload:
            # Upload scene json file
            s3_smart_upload(
                bucket=bucket,
                key=f"{path}/scene.json",
                fileobj=BytesIO(bytes(ujson.dumps(scene_dict), encoding="utf-8")),
                content_type="application/json",
            )

            # Upload ego2world csv file
            s3_smart_upload(
                bucket=bucket,
                key=f"{path}/ego2world.csv",
                fileobj=BytesIO(bytes(poses_csv, encoding="utf-8")),
                content_type="text/plain",
            )

            if use_threads:
                p = ThreadPool(processes=UPLOAD_POOL_SIZE)
                func = partial(
                    NucleusLidarFrame.s3_upload,
                    bucket=bucket,
                    path=path
                )
                p.map(func, self.frames)
            else:
                for frame in self.frames:
                    frame.s3_upload(bucket, path)

        signed_url = get_signed_url(bucket, f"{path}/scene.json")

        print(f"Scene uploaded: {signed_url}")
        return self.base_url

    def generate_nucleus_scene(
        self,
        ref_id_prefix: str = "",
        presigned_items: bool = False
    ) -> nucleus.LidarScene:
        """Generates the Nucleus Scene object that can be asynchronously uploaded to the platform

        :param ref_id_prefix: a prefix that can be added to reference ID of the scene and dataset items to ensure unique values, if not already existent
        :type ref_id_prefix: string
        :param presigned_items: Dictates that all items involved in Nucleus scene are presigned via S3
        :type presigned_items: bool

        :returns: the fully constructed Nucleus LidarScene
        :type LidarScene: nucleus.LidarScene
        """

        if ref_id_prefix is not "":
            self.ref_id_prefix = ref_id_prefix

        assert self.ref_id_prefix is not "", "Please set a Reference ID prefix to ensure reference idempotency."

        scene_dict = self.to_dict()
        nucleus_frames = []
        for frame_idx, frame in enumerate(self.frames):

            cam_items = frame.generate_cam_nucleus_dataset_items(
                scene_dict=self.to_dict(),
                ref_id_prefix=self.ref_id_prefix,
                presigned_items=presigned_items,
            )

            pointcloud_location = scene_dict["frames"][frame_idx]

            if presigned_items:
                pointcloud_location = get_signed_url(
                    bucket=urlparse(pointcloud_location).netloc,
                    path=urlparse(pointcloud_location).path[1:],
                )

            nucleus_frame = nucleus.Frame(
                lidar=nucleus.DatasetItem(
                    pointcloud_location=pointcloud_location,
                    metadata=frame.pointcloud_metadata,
                    reference_id=get_pointcloud_ref_id(self.ref_id_prefix, frame_idx)
                ),
                **cam_items,
            )

            nucleus_frames.append(nucleus_frame)

        return nucleus.LidarScene(
            reference_id=get_scene_ref_id(self.ref_id_prefix), frames=nucleus_frames
        )


import numpy as np
import transforms3d as t3d
from pyquaternion import Quaternion


class Transform:
    """Transform object represent a rigid transformation matrix (rotation and translation).

    Transform is a 4x4 matrix, although it could be instance using (16,1), (3,4), (3,3) or (3,1) matrixes.
    **Note**: not all the methods from Transform will work using scaled/small matrixes

    .. highlight:: python
    .. code-block:: python

      [
        [ r00, r01, r02, t0],
        [ r10, r11, r12, t1],
        [ r20, r21, r22, t2],
        [ 0, 0, 0, 1]
      ]

    """

    def __init__(self, value=None):
        self.matrix = np.eye(4)

        if isinstance(value, Transform):
            self.matrix = np.array(value)

        elif isinstance(value, Quaternion):
            self.rotation = value.rotation_matrix

        elif value is not None:
            value = np.array(value)
            if value.shape == (4, 4):
                self.matrix = value

            if value.shape == (16,):
                self.matrix = np.array(value).reshape((4, 4))

            elif value.shape == (3, 4):
                self.matrix[:3, :4] = value

            elif value.shape == (3, 3):
                self.rotation = value

            elif value.shape == (3,):
                self.translation = value

    @staticmethod
    def from_Rt(R, t):
        """Create a transform based on a rotation and a translation components.

        :param R: Rotation matrix or quaternion.
        :type R: Quaternion, list
        :param t: Translation component
        :type t: list
        :returns: Transform created based on the components
        :rtype: Transform

        """
        if isinstance(R, Quaternion):
            R = R.rotation_matrix
        return Transform(np.block([[R, np.mat(t).T], [np.zeros(3), 1]]))

    @staticmethod
    def from_euler(angles, axes="sxyz", degrees=False):
        """Create a transform from euler angles

        :param angles: Values of the rotation per axis
        :type angles: list
        :param axes: Order of the axis (default ``sxyz``)
        :type axes: str
        :param degrees: Use degrees or radians values (default ``False`` = radians)
        :type degrees: boolean
        :returns: Transform created from euler angles
        :rtype: Transform

        """
        if degrees:
            angles = np.deg2rad(angles)
        return Transform(t3d.euler.euler2mat(*angles, axes=axes))

    @staticmethod
    def from_transformed_points(A, B):
        """Create a transform from two points

        :param A: Point A (x,y,z)
        :type A: list
        :param B: Point B (x,y,z)
        :type B: list
        :returns: Transform created from the angles
        :rtype: Transform

        """
        assert A.shape == B.shape

        mean_A = np.mean(A, axis=0)
        mean_B = np.mean(B, axis=0)
        centroid_A = A - mean_A
        centroid_B = B - mean_B

        C = centroid_A.T @ centroid_B
        V, S, W = np.linalg.svd(C)

        if (np.linalg.det(V) * np.linalg.det(W)) < 0.0:
            S[-1] = -S[-1]
            V[:, -1] = -V[:, -1]

        R = V @ W
        t = mean_B - mean_A @ R

        return Transform.from_Rt(R.T, t)

    @staticmethod
    def random():
        """Create a transform from random rotation and translation

        :returns: Transform created based on the angles
        :rtype: Transform

        """
        return Transform.from_Rt(Quaternion.random(), np.random.rand(3))

    @property
    def rotation(self):
        """Transform rotation

        :getter: Return transform's rotation
        :setter: Set transform rotation, could use a 3x3 matrix or a Quaternion
        :type: 3x3 matrix
        """
        return self.matrix[:3, :3]

    @property
    def quaternion(self):
        """Transform rotation as quaternion

        :getter: Return transform's rotation as quaternion
        :type: Quaternion
        """
        return Quaternion(t3d.quaternions.mat2quat(self.matrix[:3, :3]))

    @rotation.setter
    def rotation(self, rotation):
        if isinstance(rotation, Quaternion):
            rotation = rotation.rotation_matrix
        self.matrix[:3, :3] = rotation

    @property
    def position(self):
        """Transform position/translation

        :getter: Return transform's position
        :setter: Set transform's position list(3x1)
        :type: list
        """
        return self.matrix[:3, 3].flatten()

    @position.setter
    def position(self, position):
        self.matrix[:3, 3] = np.array(position).reshape(3)

    @property
    def translation(self):
        """Transform position/translation

        :getter: Return transform's position
        :setter: Set transform's position list(3x1)
        :type: list
        """
        return self.matrix[:3, 3].flatten()

    @translation.setter
    def translation(self, translation):
        self.matrix[:3, 3] = np.array(translation).reshape(3)

    @property
    def euler_angles(self, axes="sxyz"):
        """Transform rotation in euler angles

        :getter: Return transform's rotaiton in euler angles
        :type: list
        """
        return t3d.euler.mat2euler(self.matrix, axes=axes)

    @property
    def euler_degrees(self, axes="sxyz"):
        """Transform rotation in euler degrees

        :getter: Return transform's rotaiton in euler degrees
        :type: list
        """
        return np.rad2deg(t3d.euler.mat2euler(self.matrix, axes=axes))

    @property
    def T(self):
        """Transpose of the transform

        :returns: Transpose of the transform
        :rtype: Transform

        """
        try:
            return Transform(self.matrix.T)
        except ValueError:
            print("Can not transpose the Transform matrix")

    @property
    def inverse(self):
        """Inverse of the transform

        :returns: Inverse of the transform
        :rtype: Transform

        """
        try:
            return Transform.from_Rt(
                self.rotation.T, np.dot(-self.rotation.T, self.translation)
            )
        except ValueError:
            print("Can not inverse the Transform matrix")

    def apply(self, points):
        """Apply transform to a list of points

        :param points: List of points (N,3) or (N,4)
        :returns: List of points witht the transform applied
        :rtype: list

        """
        points_4d = np.hstack([points[:, :3], np.ones((points.shape[0], 1))])
        transformed_4d = points_4d.dot(self.matrix.T)
        return np.hstack([transformed_4d[:, :3], points[:, 3:]])

    def interpolate(self, other, factor):
        """Interpotation of the transform

        :param other: Transform to interpolate with
        :type other: Transform
        :param factor: Factor of interpolation
        :type factor: float between 0 and 1
        :returns: Transform resulted from the interpolation
        :rtype: Transform

        """
        assert 0 <= factor <= 1.0
        other = Transform(other)
        return self.from_Rt(
            Quaternion.slerp(self.quaternion, other.quaternion, factor),
            self.position + factor * (other.position - self.position),
        )

    def __array__(self):
        return self.matrix

    def __getitem__(self, values):
        return self.matrix.__getitem__(values)

    def __add__(self, other):
        return Transform(other) @ self

    def __matmul__(self, other):
        if isinstance(other, np.ndarray):
            return self.apply(other)
        return Transform(self.matrix @ Transform(other).matrix)

    def __eq__(self, other):
        return np.allclose(self.matrix, other.matrix)

    def __repr__(self):
        return "R=%s t=%s" % (
            np.array_str(self.euler_degrees, precision=3, suppress_small=True),
            np.array_str(self.position, precision=3, suppress_small=True),
        )



import random

import cv2
import numpy as np
from PIL import Image, ImageDraw

from .color_utils import map_colors
from .transform import Transform


class LidarCamera:
    """Camera object that contains all the camera information

    Camera properties:
        - id = camera id/Name/Identifier, type: int, str
        - pose: Camera pose/extrinsic, type: Transform
        - world_poses: World poses, this will make the camera ignore the frame poses, type: list(Transform)
        - K: Intrinsic matrix
        - D: Camera distortion coefficients [k1,k2,p1,p2,k3,k4], default all set to ``0``
        - model: Camera model, default ``brown_conrady``
        - scale_factor: Camera scale factor, default ``1``
        - skew: Camera scale factor, default ``0``

    Usefull extra documentation to understand better how this object works:
    https://docs.opencv.org/3.4/d9/d0c/group__calib3d.html

    """

    world2cam = Transform.from_euler([-90, 0, -90], degrees=True)

    def __init__(self, camera_id):
        self.id = camera_id
        self.pose = Transform(self.world2cam)
        self.world_poses = None
        self.K = np.eye(3, dtype=np.float32)
        self.D = np.zeros(6, dtype=np.float32)
        self.model = "brown_conrady"
        self.scale_factor = 1
        self.skew = 0

    @property
    def position(self) -> np.ndarray:
        """Camera position

        :getter: Return camera's position
        :setter: Set camera's position
        :type: list(x,y,z)
        """
        return self.pose.position

    @property
    def rotation(self) -> np.ndarray:
        """Camera rotation/heading

        :getter: Return camera's rotation
        :setter: Set camera's rotation
        :type: 3x3 rotation matrix
        """
        return self.pose.rotation

    @property
    def world_transform(self) -> Transform:
        """World transform/pose (to avoid frame pose)

        :getter: pose @ world2cam.T
        :setter: pose = transform @ world2cam
        :type: Transform
        """
        return self.pose @ self.world2cam.T

    @property
    def fx(self):
        """Camera X focal length

        :getter: Return camera's X focal length
        :type: double
        """
        return self.K[0, 0]

    @property
    def fy(self):
        """Camera Y focal length

        :getter: Return camera's Y focal length
        :type: double
        """
        return self.K[1, 1]

    @property
    def cx(self):
        """Camera X center point

        :getter: Return camera's X center point
        :type: double
        """
        return self.K[0, 2]

    @property
    def cy(self):
        """Camera Y center point

        :getter: Return camera's Y center point
        :type: double
        """
        return self.K[1, 2]

    @property
    def intrinsic_matrix(self):
        """Camera intrinsic/K

        :getter: Return camera's intrinsic matrix
        :type: 3x3 matrix
        """
        return self.K

    @property
    def extrinsic_matrix(self):
        """Camera extrinsic

        :getter: Return camera's extrinsic matrix (pose.inverse[:3, :4])
        :setter: pose = Transform(matrix).inverse
        :type: 3x4 matrix
        """
        return self.pose.inverse[:3, :4]

    @property
    def projection_matrix(self):
        """Projection matrix

        :getter: K @ extrinsic_matrix
        :setter: K, R, t, _, _, _, _ = cv2.decomposeProjectionMatrix(projection_matrix)
        :type: 3x4 projection matrix
        """
        return self.K @ self.extrinsic_matrix

    @position.setter
    def position(self, position: np.ndarray):
        self.pose.position = position

    @rotation.setter
    def rotation(self, rotation):
        self.pose.rotation = Transform(rotation).rotation

    @world_transform.setter
    def world_transform(self, transform: Transform):
        self.pose = transform @ self.world2cam

    @extrinsic_matrix.setter
    def extrinsic_matrix(self, matrix):
        self.pose = Transform(matrix).inverse

    @projection_matrix.setter
    def projection_matrix(self, P):
        assert P.shape == (3, 4), "Projection matrix should be 3x4"
        K, R, t, _, _, _, _ = cv2.decomposeProjectionMatrix(P)
        self.pose = Transform.from_Rt(R.T, t[:3, 0] / t[3, 0])
        self.K = K

    def calibrate(
        self,
        position=None,
        rotation=None,
        pose=None,
        extrinsic_matrix=None,
        projection_matrix=None,
        K=None,
        D=None,
        model=None,
        scale_factor=None,
        skew=None,
        world_transform=None,
        world_poses=None,
        **kwargs
    ):
        """Helper for camera calibration

        Args:
          position (list(int)): Camera position [x, y, z]
          rotation (rotation matrix): Camera rotation/heading
          pose (Transform): Camera pose (position + rotation)
          extrinsic_matrix (matrix 4x4): Extrinsic 4x4 matrix (world to camera transform) (pose = Transform(matrix).inverse)
          projection_matrix (matrix 3x4): 3x4 projection matrix (K, R, t, _, _, _, _ = cv2.decomposeProjectionMatrix(projection_matrix))
          K (matrix 3x3): Intrinsic 3x3 matrix
          D (list(double)): Distortion values following this order: [k1,k2,p1,p2,k3,k4,k5,k6], required [k1,k2,p1,p2,k3,k4]
          model (str): Camera model
          scale_factor (int): Image scale_factor
          skew (int): Camera skew coefficient
          world_transform (Transform): Overwrite camera pose with the world transform (pose = transform @ world2cam)
          world_poses (list(Transform)):  World poses, this will make the camera ignore the frame poses

        Keyword Args:
          fx (str): Focal length in X
          fy (str): Focal length in Y
          cx (str): Center point in X
          cy (str): Center point in Y
          k1 (double): Distortion value k1
          k2 (double): Distortion value k2
          k3 (double): Distortion value k3
          k4 (double): Distortion value k4
          k5 (double): Distortion value k5
          k6 (double): Distortion value k6
          p1 (double): Distortion value p1
          p2 (double): Distortion value p2

        """
        if position is not None:
            self.position = position
        if rotation is not None:
            self.rotation = rotation
        if pose is not None:
            self.pose = Transform(pose)
        if extrinsic_matrix is not None:
            self.extrinsic_matrix = extrinsic_matrix
        if projection_matrix is not None:
            self.projection_matrix = projection_matrix
        if K is not None:
            self.K = np.array(K[:3, :3])
        if D is not None:
            assert (
                len(D) >= 6
            ), "Distortion list should have have at least these values [k1,k2,p1,p2,k3,k4]"
            self.D = D
        if model is not None:
            self.model = model
        if scale_factor is not None:
            self.scale_factor = scale_factor
        if skew is not None:
            self.skew = skew
        if world_transform is not None:
            self.world_transform = world_transform
        if world_poses is not None:
            self.world_poses = world_poses
        if "fx" in kwargs:
            self.K[0, 0] = kwargs["fx"]
        if "fy" in kwargs:
            self.K[1, 1] = kwargs["fy"]
        if "cx" in kwargs:
            self.K[0, 2] = kwargs["cx"]
        if "cy" in kwargs:
            self.K[1, 2] = kwargs["cy"]
        if "k1" in kwargs:
            self.D[0] = kwargs["k1"]
        if "k2" in kwargs:
            self.D[1] = kwargs["k2"]
        if "p1" in kwargs:
            self.D[2] = kwargs["p1"]
        if "p2" in kwargs:
            self.D[3] = kwargs["p2"]
        if "k3" in kwargs:
            self.D[4] = kwargs["k3"]
        if "k4" in kwargs:
            self.D[5] = kwargs["k4"]
        if "k5" in kwargs:
            self.D = np.lib.pad(self.D, (0, 1), "constant", constant_values=(0))
            self.D[6] = kwargs["k5"]
        if "k6" in kwargs:
            if len(self.D) == 6:
                self.D = np.lib.pad(self.D, (0, 2), "constant", constant_values=(0))
            elif len(self.D) == 7:
                self.D = np.lib.pad(self.D, (0, 1), "constant", constant_values=(0))
            self.D[7] = kwargs["k6"]

    def apply_transform(self, transform: Transform):
        """Apply transformation to the camera (transform @ pose)

        :param transform: Transform to apply to the object
        :type transform: Transform
        """

        self.pose = transform @ self.pose

    def rotate(self, angles, degrees=True):
        """Rotate the camera,
        (pose = Transform.from_euler(angles, degrees=degrees) @ pose)

        :param angles: Angles to rotate (x,y,z)
        :type angles: list(float)
        :param degrees: Use rad or degrees
        :type degrees: boolean

        """
        self.apply_transform(Transform.from_euler(angles, degrees=degrees))

    def translate(self, vector):
        """Move the camera,
        (pose = Transform(angles, degrees=degrees) @ pose)

        :param vector: [x,y,z]
        :type vector: list(float)
        """
        self.apply_transform(Transform(vector))

    def project_points(self, points: np.ndarray, use_distortion=False):
        """Return array of projected points based on camera calibration values

        - When ``use_distortion=True`` it uses: cv.fisheye.projectPoints(	objectPoints, rvec, tvec, K, D[, imagePoints[, alpha[, jacobian]]]	)

        :param points: list of points
        :type points: list(float)
        :param use_distortion: For fisheye/omni cameras (not necesary for cameras like Brown-Conrady)
        :type use_distortion: boolean
        """
        projected = Transform(self.projection_matrix) @ points[:, :3]

        # projected = ((points[:, :3] - self.position) @ self.rotation) @ self.intrinsic_matrix[:3, :3].T
        projected[:, 0] /= np.where(projected[:, 2] == 0, np.inf, projected[:, 2])
        projected[:, 1] /= np.where(projected[:, 2] == 0, np.inf, projected[:, 2])

        if use_distortion:
            projected[:, :2] = cv2.fisheye.projectPoints(
                objectPoints=np.array([points[:, :3]], dtype=np.float32),
                rvec=cv2.Rodrigues(self.extrinsic_matrix[:3, :3])[0],
                tvec=self.extrinsic_matrix[:3, 3],
                D=np.array(
                    [self.D[0], self.D[1], self.D[4], self.D[5]], dtype=np.float32
                ),
                K=np.array(self.K, dtype=np.float32),
                alpha=self.skew,
            )[0].reshape((-1, 2))

        return np.hstack([projected[:, :3], points[:, 3:]])

    def get_projected_image(
        self, image, points, frame_transform, color_mode="default", oversample=3
    ):
        """Return image with points projected onto it

        :param image: Camera image
        :type image: PIL.Image
        :param points: list of points/pointcloud
        :type points: list(float)
        :param frame_transform: Frame transform/pose
        :type frame_transform: Transform
        :param color_mode: Color mode, default ``default``, modes are: 'depth', 'intensity' and 'default'
        :type color_mode: str
        :param oversample: Padding on projected points, this is used to project points outside the image, it's useful for debugging, default ``3`` = 3 times the image size
        :type oversample: int
        :returns: Image with points projected
        :rtype: PIL.Image
        """
        assert image, "No image loaded."

        def crop_points(points, bounding_box):
            conditions = np.logical_and(
                points[:, :3] >= bounding_box[0], points[:, :3] < bounding_box[1]
            )
            mask = np.all(conditions, axis=1)
            return points[mask]

        im = image.get_image().convert("RGBA")
        radius = 3
        points = np.array(
            random.sample(points.tolist(), int(len(points) / 2))
        )  # reduce the number of points projected, no need to project everything
        # Project points image
        points_im = Image.new(
            "RGBA", (im.size[0] * oversample, im.size[1] * oversample)
        )

        draw = ImageDraw.Draw(points_im)

        if self.model == "cylindrical":
            # handle cylindrical cameras
            epsilon = 0.0000001
            fisheye = frame_transform.inverse @ np.array(
                [points[:, 0], points[:, 1], points[:, 2]], dtype=np.float
            )  # 3D point in camera coordinates
            fisheye = fisheye.T  # 3D point in camera coordinates
            fisheye = self.pose.inverse @ fisheye
            fisheye[
                :, 1
            ] *= -1  # invert y because cylindrical y is up and cartesian y is down
            fisheye = (
                Transform(self.extrinsic_matrix[:3, :3]) @ fisheye.T
            )  # lift cylinder to stand up straight
            cylindrical = np.array(
                [
                    np.arctan2(fisheye[0, :], fisheye[2, :]),
                    fisheye[1, :] / np.sqrt(fisheye[0, :] ** 2 + fisheye[2, :] ** 2),
                    np.ones(fisheye.shape[1]),
                ]
            )
            cylindrical[
                1, :
            ] *= -1  # invert y because cylindrical y is up and cartesian y is down
            q = self.K @ cylindrical
            img_coords = q[[0, 1], :]  # pixels on image
            img_coords = img_coords.T

            for point in img_coords[:, :2]:
                draw.ellipse(
                    [tuple(point - radius), tuple(point + radius)],
                    fill=tuple([255, 10, 10]),
                )
            points_im = points_im.resize(im.size, Image.CUBIC)

            # Merge images
            projected_im = Image.composite(points_im, im, points_im)
            return projected_im
        if self.model == "fisheye":
            wct = self.pose @ frame_transform.T
            projected = self.project_points(points, use_distortion=True)
            #  projected = crop_points(projected,
            #  np.array([[0, 0, 0.1], [im.size[0], im.size[1], np.inf]]))
            fisheye = self.world_transform @ np.array(
                [projected[:, 0], projected[:, 1], projected[:, 2]], dtype=np.float
            )  # 3D point in camera coordinates
            fisheye = fisheye.T  # 3D point in camera coordinates
            fisheye = np.concatenate(
                (np.array(fisheye), np.array(projected[:, 3])[:, None]), axis=1
            )

            if not len(fisheye):
                return im

            colors = map_colors(fisheye, color_mode)

            for point, color in zip(fisheye[:, :2] * oversample, colors):
                draw.ellipse(
                    [tuple(point - radius), tuple(point + radius)], fill=tuple(color)
                )
            points_im = points_im.resize(im.size, Image.CUBIC)

            # Merge images
            projected_im = Image.composite(points_im, im, points_im)
            return projected_im
        else:
            projected = self.project_points(points, use_distortion=False)
            projected = crop_points(
                projected, np.array([[0, 0, 0.1], [im.size[0], im.size[1], np.inf]])
            )

            # Returns original image if not projected points on image
            if not len(projected):
                return im

            colors = map_colors(projected, color_mode)

            for point, color in zip(projected[:, :2] * oversample, colors):
                draw.ellipse(
                    [tuple(point - radius), tuple(point + radius)], fill=tuple(color)
                )
            points_im = points_im.resize(im.size, Image.CUBIC)

            # Merge images
            projected_im = Image.composite(points_im, im, points_im)
            return projected_im

        def __repr__(self):
            return "LidarCamera({0}) {1}".format(self.id, self.pose)



import os
import hashlib
from io import BytesIO
import boto3
import ujson
import numpy as np
from scaleapi import ScaleClient

_S3 = None

# Format/ Parsing helpers


def format_lidar_point(p):
    return dict(zip(("x", "y", "z", "i", "d"), p))


def format_point(p):
    return dict(zip(("x", "y", "z"), p))


def format_quaternion(q):
    return dict(zip(["w", "x", "y", "z"], q))


def parse_xyz(obj):
    return np.array([obj["x"], obj["y"], obj["z"]])


def parse_quaternion(obj):
    return np.array([obj["w"], obj["x"], obj["y"], obj["z"]])


def parse_cuboid_frame(frame):
    return np.array([parse_xyz(p) for p in [c["position"] for c in frame["cuboids"]]])


def read_cuboid_points(filename):
    # Read from result file
    with open(filename, "r") as json_file:
        return np.array([parse_cuboid_frame(frame) for frame in ujson.load(json_file)])


# Task creation helpers


def get_api_client() -> ScaleClient:
    assert os.getenv(
        "SCALE_API_KEY"
    ), "No API key, please set SCALE_API_KEY environment variable"
    return ScaleClient(os.getenv("SCALE_API_KEY"))


def get_default_template():
    return {"task_id": None}


# S3 helpers


def get_s3_client():
    global _S3
    if _S3 is None:
        _S3 = boto3.client("s3")
    return _S3


def fp_md5(fp: BytesIO):
    fp.seek(0)
    md5 = hashlib.md5()
    buf = fp.read(2 ** 20)
    while buf:
        md5.update(buf)
        buf = fp.read(2 ** 20)
    fp.seek(0)
    return md5.hexdigest()


def get_s3_etag(bucket, key):
    s3 = get_s3_client()
    try:
        return s3.head_object(Bucket=bucket, Key=key)["ETag"].strip('"')
    except Exception:
        pass


def get_signed_url(bucket: str, path: str) -> str:
    s3 = get_s3_client()
    signed_url = s3.generate_presigned_url(
        "get_object", Params={"Bucket": bucket, "Key": path}
    )
    return signed_url


def s3_smart_upload(bucket, key, fileobj, content_type):
    s3 = get_s3_client()

    s3_hash = get_s3_etag(bucket, key)
    local_hash = fp_md5(fileobj)

    if s3_hash == local_hash:
        print(f"File exists: {bucket}/{key}")
        return

    print(f"Uploading {bucket}/{key}...")
    s3.upload_fileobj(
        Fileobj=fileobj, Bucket=bucket, Key=key, ExtraArgs={"ContentType": content_type}
    )


def scale_file_upload(fileobj, project_name):
    my_file = get_api_client().upload_file(file=fileobj, project_name=project_name,)
    return my_file.attachment_url



import cv2
import numpy as np
from matplotlib.colors import hsv_to_rgb


def normalize(values):
    return np.interp(
        values, [np.percentile(values, 5), np.percentile(values, 95)], [0, 1]
    )


def map_colors(points, mode="default", color_map=cv2.COLORMAP_JET):
    """Set point's colors based on mode

    params:
        - points: Pointcloud
        - mode: Color mode ``default``
    """

    if mode == "depth":
        values = np.uint8(normalize(points[:, 2]) * 255)
        rgb = cv2.applyColorMap(values, color_map)
        return rgb.reshape((-1, 3))

    if mode == "intensity":
        values = 255 - np.uint8(normalize(points[:, 3]) * 255)
        rgb = cv2.applyColorMap(values, color_map)
        return rgb.reshape((-1, 3))

    if mode == "default":
        x, y, z, i = points.T
        i_range = [np.percentile(i, 10), np.percentile(i, 90)]

        value = np.sum(
            [np.interp(z, [0, 50], [100, -10]), np.interp(i, i_range, [-10, 10])],
            axis=0,
        )

        rgb = hsv_to_rgb(
            np.array(
                [
                    np.interp(value, [0, 100], [220 / 360, 1 / 360]),
                    np.interp(i, i_range, [0.5, 1]),
                    np.interp(i, i_range, [0.4, 1]),
                ]
            ).T
        )

        alpha = np.interp(i, [np.percentile(i, 10), np.percentile(i, 90)], [0.75, 1])

        return np.uint8(np.hstack([rgb.reshape((-1, 3)), alpha.reshape((-1, 1))]) * 255)

    # other mode
    values = np.uint8(normalize(points[:, 2]) * 255)
    rgb = cv2.applyColorMap(values, color_map)
    alpha = np.uint8(normalize(points[:, 3]) * 255)
    return np.hstack([rgb.reshape((-1, 3)), alpha.reshape((-1, 1))])



import zipfile
from multiprocessing.pool import ThreadPool
from functools import partial
from io import BytesIO
from typing import MutableMapping, List, Dict
from tqdm import tqdm

import numpy as np
import pandas as pd
import ujson
from scaleapi.tasks import Task, TaskType

from .camera import LidarCamera
from .image import LidarImage
from .frame import LidarFrame
from .connectors import Importer
from .transform import Transform
from .helper import s3_smart_upload, get_signed_url

UPLOAD_POOL_SIZE = 8


class LidarScene:
    """LidarScene object representing all frames in a scene.

    Scene properties:
      - cameras: List of cameras
      - frames: List of frames
      - base_url: Url used to host the data in S3

    """

    def __init__(self):
        """
        :rtype: object
        """
        self.cameras: MutableMapping[LidarCamera] = pd.Series()
        self.frames: MutableMapping[LidarFrame] = pd.Series()
        self.base_url = None
        self.scale_file_attachments = None

    def get_camera(self, camera_id=None, index: int = None) -> LidarCamera:
        """Get a camera by id (or index) or create one if it does not exist

        :param camera_id: The camera id
        :type camera_id: str, int
        :param index: The camera index
        :type index: int

        :return: LidarCamera
        :rtype:  LidarCamera

        """
        assert (
            camera_id is not None or index is not None
        ), "id or index must be specified"

        if camera_id is None:
            camera_id = self.cameras.index[index]

        if camera_id not in self.cameras:
            if isinstance(camera_id, int):
                self.cameras.index = self.cameras.index.astype(int)
            self.cameras[camera_id] = LidarCamera(camera_id)
        return self.cameras[camera_id]

    def get_frame(self, frame_id=None, index: int = None) -> LidarFrame:
        """Get a frame by id (or index) or create one if it does not exist

        :param frame_id: The frame id
        :type frame_id: str, int
        :param index: The frame index
        :type index: int

        :return: LidarFrame
        :rtype: LidarFrame

        """
        assert (
            frame_id is not None or index is not None
        ), "id or index must be specified"

        if frame_id is None:
            frame_id = self.frames.index[index]

        if frame_id not in self.frames:
            if isinstance(frame_id, int):
                self.frames.index = self.frames.index.astype(int)
            self.frames[frame_id] = LidarFrame(frame_id, cameras=self.cameras)
        return self.frames[frame_id]

    def apply_transforms(self, world_transforms: List[Transform]):
        """Apply transformations to all the frames (the number of Transformation should match the number of frames)

        :param world_transforms: List of Transform
        :type world_transforms: list(Transform)

        """
        assert len(world_transforms) != len(
            self.frames
        ), "world_transforms should have the same length as frames"

        for idx in range(len(self.frames)):
            self.get_frame(index=idx).apply_transform(world_transforms[idx])

    def filter_points(self, min_intensity=None, min_intensity_percentile=None):
        """Filter points based on intensity

        :param min_intensity: Minimun intensity allowed
        :type min_intensity: int
        :param min_intensity_percentile: Minimun percentile allowed (use np.percentile)
        :type min_intensity_percentile: int

        """
        for frame in self.frames:
            frame.filter_points(min_intensity, min_intensity_percentile)

    def get_projected_image(
        self, camera_id, color_mode="intensity", frames_index=range(0, 1), **kwargs
    ):
        """Get camera_id image with projected points, (**Legacy method**)

        :param camera_id: Camera id/Name/Identifier
        :type camera_id: str, int
        :param color_mode:  Color mode, default ``default``, modes are: 'depth', 'intensity' and 'default'
        :type color_mode: str
        :param frames_index: Project points for a range of frames, default `first frame`
        :type frames_index: range

        :returns: Image with points projected
        :rtype: Image

        """
        all_points = np.array([]).reshape(0, 4)
        for idx in frames_index:
            points = np.array(self.frames[idx].points)[:, :4]
            print(points.shape)
            points[:, 3] = idx
            all_points = np.concatenate((all_points, points), axis=0)
        return self.cameras[camera_id].get_projected_image(
            self.frames[frames_index[0]].get_image(camera_id),
            all_points,
            self.frames[frames_index[0]].transform,
            color_mode,
        )

    def apply_transform(self, world_transform: Transform):
        """Apply a Transformation to all the frames

        :param world_transform: Transform to apply to all the frames
        :type world_transform: Transform

        """

        for idx in range(len(self.frames)):
            self.get_frame(index=idx).apply_transform(world_transform)

    def make_transforms_relative(self):
        """Make all the frame transform relative to the first transform/frame. This will set the first transform to position (0,0,0) and heading (1,0,0,0)"""
        offset = self.get_frame(index=0).transform.inverse
        for frame in self.frames:
            frame.transform = offset @ frame.transform

    def downsample_scene(self, voxel_size_mm: int = 250):
        """Downsamples all frames according to voxel size"""
        for idx in range(len(self.frames)):
            self.get_frame(index=idx).downsample_frame(voxel_size_mm=voxel_size_mm)

    def to_dict(self, base_url: str = None) -> dict:
        """Return a dictionary with the frame urls using the base_url as base.

        :param base_url: This url will be concatenated with the frames name, e.g.: `'%s/frame-%s.json' % (base_url, frame.id)`
        :type base_url: str
        :return: Dictionary with the frame urls data
        :rtype: dict

        """
        if base_url is None:
            base_url = self.base_url
        return dict(
            frames=["%s/frame-%s.json" % (base_url, frame.id) for frame in self.frames]
        )

    def s3_upload(
        self,
        bucket: str,
        path=None,
        mock_upload: float = False,
        use_threads: float = True,
    ):
        """Save scene in S3

        :param bucket: S3 Bucket name
        :type bucket: str
        :param path: Path to store data
        :type key: str
        :param mock_upload: To avoid upload the data to S3  (defualt ``False``)
        :type mock_upload: float
        :param use_threads: In order to upload multiple files at the same time using threads  (defualt ``True``)
        :type use_threads: float

        :return: Scene S3 url
        :rtype: str

        """
        self.base_url = f"s3://{bucket}/{path}"

        print("Uploading scene to S3: %s" % self.base_url)
        scene_dict = self.to_dict(self.base_url)

        poses_csv = pd.DataFrame(
            self.frames.map(lambda f: list(f.transform.matrix.reshape(-1))).to_dict()
        ).T.to_csv(header=False)

        if not mock_upload:
            # Upload scene json file
            s3_smart_upload(
                bucket=bucket,
                key=f"{path}/scene.json",
                fileobj=BytesIO(bytes(ujson.dumps(scene_dict), encoding="utf-8")),
                content_type="application/json",
            )

            # Upload ego2world csv file
            s3_smart_upload(
                bucket=bucket,
                key=f"{path}/ego2world.csv",
                fileobj=BytesIO(bytes(poses_csv, encoding="utf-8")),
                content_type="text/plain",
            )

            if use_threads:
                p = ThreadPool(processes=UPLOAD_POOL_SIZE)
                func = partial(LidarFrame.s3_upload, bucket=bucket, path=path)
                p.map(func, self.frames)
            else:
                for frame in self.frames:
                    frame.s3_upload(bucket, path)

        signed_url = get_signed_url(bucket, f"{path}/scene.json")

        print(f"Scene uploaded: {signed_url}")
        return self.base_url

    def scale_file_upload(
        self, project_name: str,
    ):
        """Save scene in Scale file

        :param project_name: File project name
        :type project_name: str
        :param verbose: Set to false to not show the progress bar

        :return: Scene file url
        :rtype: str

        """

        print("Uploading scene to Scale file")
        p = ThreadPool(processes=UPLOAD_POOL_SIZE)
        func = partial(LidarFrame.scale_file_upload, project_name=project_name)
        self.scale_file_attachments = p.map(func, self.frames)

        print(
            f"Finishes uploading scene to Scale file, uploaded {len(self.scale_file_attachments)} frames"
        )
        return self.scale_file_attachments

    def save_task(self, filepath: str, template=None):
        """Save the entire scene (with frame and images) in zipfile format to local filepath

        :param filepath: File name and path in which the scene should be saved
        :type filepath: str

        """
        print("Saving scene:", filepath)
        with zipfile.ZipFile(filepath, mode="w") as out:
            # Save task
            scene_dict = self.to_dict()
            task_dict = dict(
                template or {}, attachment_type="json", attachments=scene_dict["frames"]
            )
            out.writestr("task.json", ujson.dumps(task_dict))

            # Save frames
            for frame in self.frames:
                # Save points
                data = frame.to_json(self.base_url)
                out.writestr(
                    "frame-%s.json" % frame.id, data, compress_type=zipfile.ZIP_DEFLATED
                )

                # Save frame images
                for camera_id, image in frame.images.items():
                    if not image.image_path:
                        assert NotImplementedError(
                            "Only file-imported Images supported"
                        )
                    out.write(
                        image.image_path, "image-%s-%s.jpg" % (camera_id, frame.id)
                    )
        print("Scene saved.")

    def create_task(
        self, template: Dict = None, task_type: TaskType = TaskType.LidarAnnotation
    ) -> Task:
        """Create a Scale platform task from the configured scene

        :param template: Dictionary of payload for task creation (https://private-docs.scale.com/?python#parameters), attachments data will be filled automatically.
        :type template: dict
        :param task_type: Select a Scale API endpoint top upload data to, currently supports 'lidarannotation', 'lidarsegmentation', and 'lidartopdown'. Defaults to 'lidarannotation'.
        :type task_type: str

        :return: Task object with related information. Inherited from `scaleapi.Task` object.
        :rtype: Task
        """
        if task_type == TaskType.LidarAnnotation:
            from .task import LidarAnnotationTask

            return LidarAnnotationTask.from_scene(self, template)
        elif task_type == TaskType.LidarTopdown:
            from .task import LidarTopDownTask

            return LidarTopDownTask.from_scene(self, template)
        elif task_type == TaskType.LidarSegmentation:
            from .task import LidarSegmentationTask

            return LidarSegmentationTask.from_scene(self, template)
        else:
            raise NotImplementedError(
                f"Specified task_type {task_type} is not supported"
            )



import numpy as np
import requests
from scaleapi.tasks import Task, TaskType
import ujson

from .scene import LidarScene
from .helper import parse_xyz, get_api_client, get_default_template


class LidarAnnotationTask(Task):
    """Lidar annotation Task object"""

    scene: LidarScene = None

    def __init__(self, param_dict, client):
        super(LidarAnnotationTask, self).__init__(param_dict, client)

    @staticmethod
    def from_scene(scene: LidarScene, template=None, client=None):
        """Load scene data and convert it into a LidarAnnotation Task format

        :param scene: Scene to load
        :type scene: LidarScene
        :param template: Template/payload to use get fetch the Scale API
        :type template: dict
        :param client: ScaleClient object, by default it will load your SCALE_API_KEY from you env vars and set a client automatically.
        :type client: scaleapi.ScaleClient
        :returns: LidarAnnotationTask object
        :rtype: LidarAnnotationTask

        """
        if not scene.scale_file_attachments:  # s3 upload
            assert (
                scene.base_url
            ), "No public URL on scene, please upload or save the scene first"

        # Get client using environ api key
        if client is None:
            client = get_api_client()

        param_dict = get_default_template()

        # Load scene params
        if not scene.scale_file_attachments:  # s3 upload
            scene_dict = scene.to_dict(scene.base_url)
            param_dict["attachments"] = scene_dict["frames"]
        else:  # scale file upload
            param_dict["attachments"] = scene.scale_file_attachments

        param_dict["attachment_type"] = "json"

        if isinstance(template, dict):
            param_dict.update(template)

        elif isinstance(template, str):
            param_dict.update(ujson.load(open(template)))

        elif template is not None:
            raise AttributeError("Template error")

        return LidarAnnotationTask(param_dict, client)

    @staticmethod
    def from_id(task_id: str):
        """Get LidarAnnotation task from a task id

        :param task_id: Task id
        :type task_id: str
        :returns: LidarAnnotationTask object created based on the task id data
        :rtype: LidarAnnotationTask

        """
        task = get_api_client().fetch_task(task_id)
        return LidarAnnotationTask(task.param_dict, task.client)

    def get_annotations(self):
        """Get annotations/response from a completed LidarAnnotation task

        :returns: Annotations
        :rtype: dict

        """
        assert "response" in self.param_dict, "Task without response"
        url = self.param_dict["response"]["annotations"]["url"]
        response = requests.get(url)
        return ujson.loads(response.text)

    def get_cuboid_positions_by_frame(self):
        """Get a list of each cuboid position in each frames (from a completed task)

        :returns: List of cuboids positions
        :rtype: list

        """
        annotations = self.get_annotations()
        return np.array(
            [
                np.array(
                    [parse_xyz(p) for p in [c["position"] for c in frame["cuboids"]]]
                )
                for frame in annotations
            ]
        )

    def publish(self, task_type: TaskType = TaskType.LidarAnnotation):
        """Publish/create a task, request Scale API with the LidarAnnotation data

        :param task_type: Task type to create, default ``lidarannotation``
        :rtype task_type: scaleapi.tasks.TaskType
        :returns: Task object creation from the response of the API call
        :rtype: scaleapi.tasks.Task

        """
        task = self._client.create_task(task_type, **self.as_dict())
        print("Task created: %s" % task)
        return task


class LidarTopDownTask(Task):
    """Lidar top-down Task object"""

    scene: LidarScene = None

    def __init__(self, param_dict, client):
        super(LidarTopDownTask, self).__init__(param_dict, client)

    @staticmethod
    def from_scene(scene: LidarScene, template=None, client=None):
        """Load scene data and convert it into a LidarTopDown Task format

        :param scene: Scene to load
        :type scene: LidarScene
        :param template: Template/payload to use get fetch the Scale API
        :type template: dict
        :param client: ScaleClient object, by default it will load your SCALE_API_KEY from you env vars and set a client automatically.
        :type client: scaleapi.ScaleClient
        :returns: LidarTopDownTask object
        :rtype: LidarTopDownTask

        """
        assert (
            scene.base_url
        ), "No public URL on scene, please upload or save the scene first"

        # Get client using environ api key
        if client is None:
            client = get_api_client()

        param_dict = get_default_template()

        # Load scene params
        if not scene.scale_file_attachments:  # s3 upload
            scene_dict = scene.to_dict(scene.base_url)
            param_dict["attachments"] = scene_dict["frames"]
        else:  # scale file upload
            param_dict["attachments"] = scene.scale_file_attachments
        param_dict["attachment_type"] = "json"

        if isinstance(template, dict):
            param_dict.update(template)

        elif isinstance(template, str):
            param_dict.update(ujson.load(open(template)))

        elif template is not None:
            raise AttributeError("Template error")

        return LidarTopDownTask(param_dict, client)

    @staticmethod
    def from_id(task_id: str):
        """Get LidarTopDown task from a task id

        :param task_id: Task id
        :type task_id: str
        :returns: LidarTopDownTask object created based on the task id data
        :rtype: LidarTopDownTask

        """
        task = get_api_client().fetch_task(task_id)
        return LidarTopDownTask(task.param_dict, task.client)

    def publish(self, task_type: TaskType = TaskType.LidarTopdown):
        """Publish/create a task, request Scale API with the LidarTopDown data

        :param task_type: Task type to create, default ``lidartopdown``
        :rtype task_type: scaleapi.tasks.TaskType
        :returns: Task object creation from the response of the API call
        :rtype: scaleapi.tasks.Task

        """
        task = self._client.create_task(task_type, **self.as_dict())
        print("Task created: %s" % task)
        return task


class LidarSegmentationTask(Task):
    """Lidar segmentation Task object"""

    scene: LidarScene = None

    def __init__(self, param_dict, client):
        super(LidarSegmentationTask, self).__init__(param_dict, client)

    @staticmethod
    def from_scene(scene: LidarScene, template=None, client=None):
        """Load scene data and convert it into a LidarSegmentation Task format

        :param scene: Scene to load
        :type scene: LidarScene
        :param template: Template/payload to use get fetch the Scale API
        :type template: dict
        :param client: ScaleClient object, by default it will load your SCALE_API_KEY from you env vars and set a client automatically.
        :type client: scaleapi.ScaleClient
        :returns: LidarSegmentationTask object
        :rtype: LidarSegmentationTask

        """
        assert (
            scene.base_url
        ), "No public URL on scene, please upload or save the scene first"

        # Get client using environ api key
        if client is None:
            client = get_api_client()

        param_dict = get_default_template()

        # Load scene params
        if not scene.scale_file_attachments:  # s3 upload
            scene_dict = scene.to_dict(scene.base_url)
            param_dict["attachments"] = scene_dict["frames"]
        else:  # scale file upload
            param_dict["attachments"] = scene.scale_file_attachments
        param_dict["attachment_type"] = "json"

        if isinstance(template, dict):
            param_dict.update(template)

        elif isinstance(template, str):
            param_dict.update(ujson.load(open(template)))

        elif template is not None:
            raise AttributeError("Template error")

        return LidarSegmentationTask(param_dict, client)

    @staticmethod
    def from_id(task_id: str):
        """Get LidarSegmentation task from a task id

        :param task_id: Task id
        :type task_id: str
        :returns: LidarSegmentationTask object created based on the task id data
        :rtype: LidarSegmentationTask

        """
        task = get_api_client().fetch_task(task_id)
        return LidarSegmentationTask(task.param_dict, task.client)

    def publish(self, task_type: TaskType = TaskType.LidarSegmentation):
        """Publish/create a task, request Scale API with the LidarSegmentation data

        :param task_type: Task type to create, default ``lidarsegmentation``
        :rtype task_type: scaleapi.tasks.TaskType
        :returns: Task object creation from the response of the API call
        :rtype: scaleapi.tasks.Task

        """
        task = self._client.create_task(task_type, **self.as_dict())
        print("Task created: %s" % task)
        return task



import shutil
import numpy as np
import tempfile
from PIL import Image, ImageEnhance
from .helper import s3_smart_upload, scale_file_upload


class LidarImage:
    """LidarImage objects represent an image with a LidarCamera reference.

    LidarImage properties:
      - camera: Camera id
      - image_path: Image path
      - transform: Transformation apply to LidarImage (will be used as: LidarImage.transform or LidarFrame.transform) @ camera.pose)
      - metadata: Metadata related to the image
      - timestamp: Timestamp
    """

    def __init__(self, camera):
        self.camera = camera
        self.image_path = None
        self.transform = None
        self.metadata = None
        self.timestamp = None

    # Legacy method
    def load_file(self, file: str):
        """Set LidarImage image_path
        (**Legacy method**)

        :param file: Set image path
        :type file: str
        """
        if not isinstance(file, str):
            print("WARNING: No file!")
        self.image_path = file

    def save_pil_image(self, pil_image: Image.Image):
        """Save image in image_path

        :param pil_image: Image to save
        :type pil_image: PIL.Image
        """
        self.image_path = tempfile.mktemp(suffix="jpg")
        pil_image = pil_image.convert("RGB")
        pil_image.save(self.image_path, format="JPEG", quality=70, optimize=True)
        print(f"Temp file created: {self.image_path}")

    def get_image(self) -> Image:
        """Open LidarImage

        :return: Image.open
        """
        return Image.open(self.image_path)

    def as_array(self) -> np.asarray:
        """Get the image as numpy array

        :returns: image as numpy array
        :rtype: np.asarray

        """
        return np.asarray(self.get_image())

    def set_scale(self, scale_factor: float):
        """Change image scale and save in image_path

        :param scale_factor: Scale factor
        :type scale_factor: float

        """
        im = self.get_image()
        size = (int(im.width * scale_factor), int(im.height * scale_factor))
        self.save_pil_image(im.resize(size, Image.LANCZOS))

    def set_brightness(self, factor: float):
        """Change image brightness and save in image_path
        (will use PIL.ImageEnhance.Brightness)


        :param factor: Brightness factor
        :type scale_factor: float

        """
        im = ImageEnhance.Brightness(self.get_image()).enhance(factor)
        self.save_pil_image(im)

    def save(self, target_file: str):
        """Save image in target_file path

        :param target_file: Path in which the image should be saved
        :type target_file: str

        """
        if not isinstance(target_file, str):
            print("WARNING: No file path!")
        shutil.copyfile(self.image_path, target_file)

    def s3_upload(self, bucket: str, key: str):
        """Save image in S3

        :param bucket: S3 Bucket name
        :type bucket: str
        :param key: file name
        :type key: str

        """
        with open(self.image_path, "rb") as fp:
            s3_smart_upload(
                bucket=bucket, key=key, fileobj=fp, content_type="image/jpeg"
            )

    def scale_file_upload(self, project_name: str):
        """Save image in Scale File

        :param project_name: File project name
        :type bucket: str
        """
        with open(self.image_path, "rb") as fp:
            return scale_file_upload(fp, project_name)



from abc import ABCMeta, abstractmethod

import numpy as np
import open3d as o3d
import pandas as pd
from laspy.file import File

_FIELDS = ["x", "y", "z", "i", "d"]


class Importer:
    """Points importer/helper"""

    class Base(metaclass=ABCMeta):
        """Abstract importer class to be inherited for data type specific implementation"""

        def __init__(self, fp: str) -> None:
            """Constructor to be called for preparing filepaths

            :param fp: Relative or absolute path to file to be loaded in explicit `load` method
            :type fp: str
            """
            self._filepath: str = fp
            self._data: pd.DataFrame = pd.DataFrame(columns=_FIELDS)

        def load(self, **kwargs) -> None:
            self._load(**kwargs)

        @property
        def data(self):
            return self._data.to_numpy()

        @abstractmethod
        def _load(self, **kwargs) -> None:
            ...

    class CSV(Base, metaclass=ABCMeta):
        """Metaclass for specific CSV Importer implementations.
        Due to non-standardized csv format, it is recommended to write more specific implementations on a case-by-case basis.
        """

        @abstractmethod
        def _load(self, **kwargs):
            ...

    class OrderedCSV(CSV):
        """Expects a csv file with or without header, but assumes correct order of columns / values as follows: `[x, y, z, i, d]`
        Cuts off any exceeding column count.
        """

        def _load(self, **kwargs):
            self._data = pd.read_csv(self._filepath, **kwargs)
            c_count = len(self._data.columns)
            if c_count > 5:  # Remove columns exceeding count of FIELDS
                self._data = self._data.drop(self._data.columns[5:], axis=1)
                c_count = 5

            self._data.columns = _FIELDS[:c_count]

    class NamedCSV(CSV):
        """Expects a csv file with header row and column names matching `x`, `y`, `z`, `i`, `d`
        Dismisses any columns not matching any of the expected names. Case-sensitive.
        """

        def _load(self, **kwargs):
            self._data = pd.read_csv(self._filepath, **kwargs)
            exists = []
            for c in _FIELDS:
                if c in self._data.columns:
                    exists.append(c)
            self._data = self._data[exists]

    class PCD(Base):
        """Uses open3d library to read pcd files, expects `[x, y, z]`, dismisses other columns."""

        def _load(self, **kwargs):
            self._data = pd.DataFrame(
                np.asarray(o3d.io.read_point_cloud(self._filepath).points)[:, :3],
                columns=_FIELDS[:3],
            )

    class LAS(Base):
        """Uses Laspy library to read .las file and expects properties `x`, `y`, `z`, `intensity` to be present."""

        def _load(self, **kwargs):
            las_file = File(self._filepath, mode="r")
            self._data = pd.DataFrame(
                np.column_stack(
                    [las_file.x, las_file.y, las_file.z, las_file.intensity]
                ),
                columns=["x", "y", "z", "i"],
            )



import numpy as np
import pandas as pd

from .transform import Transform


def interpolate_transforms(data: pd.Series) -> pd.Series:
    # This is a mask of the elements that are NaN
    empty = data.isnull().values

    # Position and indexes of empty and valid values
    valid_loc = np.argwhere(~empty).squeeze(axis=-1)
    empty_loc = np.argwhere(empty).squeeze(axis=-1)
    valid_index = data.index[valid_loc].values
    empty_index = data.index[empty_loc].values

    # Missing values before first or after last valid are discarded
    empty_loc = empty_loc[(empty_loc > valid_loc.min()) & (empty_loc < valid_loc.max())]

    # Important bit! This tells you the which valid values must be used as interpolation ends for each missing value
    interp_loc_end = np.searchsorted(valid_loc, empty_loc)
    interp_loc_start = interp_loc_end - 1

    # Values to interpolate
    valid_transforms = data.values[valid_loc]

    for i in range(len(empty_loc)):
        transform_a: Transform = valid_transforms[interp_loc_start[i]]
        transform_b: Transform = valid_transforms[interp_loc_end[i]]
        index_a = valid_index[interp_loc_start[i]]
        index_b = valid_index[interp_loc_end[i]]
        factor = (empty_index[i] - index_a) / (index_b - index_a)
        current_transform = transform_a.interpolate(transform_b, factor)
        data.iloc[empty_loc[i]] = current_transform

    return data



import unittest
from pyquaternion import Quaternion
import numpy as np
from scale_lidar_io import LidarScene, Transform


class LidarTest(unittest.TestCase):
    """Test lidar frame implementation.
    The goal is to test if the frame poses working correctly
    """

    scene = LidarScene()

    def setUp(self):
        # Setup poses
        poses_data = {
            "heading": [
                0.9992067938556252,
                0.0,
                0.0,
                0.0398218923804761,
            ],
            "position": [
                0.0,
                0.0,
                0.0,
            ],
        }

        poses = []
        pose = Transform.from_Rt(
            R=Quaternion(
                poses_data["heading"][0],
                poses_data["heading"][1],
                poses_data["heading"][2],
                poses_data["heading"][3],
            ),
            t=np.array(
                [
                    poses_data["position"][0],
                    poses_data["position"][1],
                    poses_data["position"][2],
                ]
            ),
        )
        poses.append(pose)

        lidar_extrinsics = Transform.from_Rt(
            R=Quaternion(
                1.0,
                0.0,
                0.0,
                0.0,
            ).rotation_matrix,
            t=[-3.0, 0, 2.8572714],
        )

        points = np.array(
            [
                [6.592881, 7.3447638, -1.0611035, 181.3016],
                [6.0755334, 7.2042847, -0.9519308, 103.005844],
                [8.022584, 11.820594, 0.24009487, 161.13264],
                [7.4657607, 11.8752, 0.31154653, 169.24597],
            ]
        )

        self.scene.get_frame(0).add_points(points, transform=lidar_extrinsics)
        self.scene.get_frame(0).transform = poses[0]

    def test_poses(self):
        np.testing.assert_allclose(
            self.scene.get_frame(0).transform.position, [0.0, 0.0, 0.0]
        )
        np.testing.assert_allclose(
            self.scene.get_frame(0).transform.rotation,
            [
                [0.99682843, -0.07958061, 0.0],
                [0.07958061, 0.99682843, 0.0],
                [0.0, 0.0, 1.0],
            ],
        )

    def test_points(self):
        np.testing.assert_allclose(
            self.scene.get_frame(0).points[0],
            [3.5928812, 7.34476376, 1.79616794, 181.30160522, 0.0],
        )


class CameraTest(unittest.TestCase):
    """Test cameras implementation.
    The goal is to test if the cameras extrinsic
    and intrinsic are working correctly
    """

    scene = LidarScene()
    cams = {"forward": {}}

    def setUp(self):
        # Setup cameras
        extr = {
            "heading": [
                0.9992307979502907,
                0.0009624406647958967,
                0.03318256034771476,
                0.020875914934737397,
            ],
            "position": [
                -2.1498306,
                0.00635,
                2.346325,
            ],
        }

        intr = {
            "fx": -338.5609808656974,
            "fy": -335.3931513851273,
            "cx": -638.9593048466166,
            "cy": -397.69312118769875,
            "K": [
                [338.5609808656974, 0.0, 638.9593048466166],
                [0.0, 335.3931513851273, 397.69312118769875],
                [0.0, 0.0, 1.0],
            ],
            "D": [
                0.08577354267758332,
                -0.009275696593772212,
                0.002280136164042399,
                -0.0008592168095191392,
            ],
        }

        K = np.array(intr["K"])
        D = intr["D"]

        world2optical = Transform.from_euler([-90, 0, -90], degrees=True)
        extr_R = Quaternion(extr["heading"]).rotation_matrix @ world2optical.rotation
        extr_t = extr["position"]
        extr_opticalcoord = Transform.from_Rt(R=extr_R, t=extr_t)
        self.scene.get_camera("forward").calibrate(
            extrinsic_matrix=extr_opticalcoord.inverse,
            K=K,
            D=[D[0], D[1], 0.0, 0.0, D[2], D[3]],
            model="fisheye",
        )
        # Setup poses
        poses_data = {
            "heading": [
                0.9992067938556252,
                0.0,
                0.0,
                0.0398218923804761,
            ],
            "position": [
                0.0,
                0.0,
                0.0,
            ],
        }

        poses = []
        pose = Transform.from_Rt(
            R=Quaternion(
                poses_data["heading"][0],
                poses_data["heading"][1],
                poses_data["heading"][2],
                poses_data["heading"][3],
            ),
            t=np.array(
                [
                    poses_data["position"][0],
                    poses_data["position"][1],
                    poses_data["position"][2],
                ]
            ),
        )
        poses.append(pose)

        self.scene.get_frame(0).transform = poses[0]

    def test_camera_extrinsic(self):
        np.testing.assert_allclose(
            self.scene.get_frame(0).cameras["forward"].pose.position,
            [-2.1498306, 0.00635, 2.346325],
        )
        np.testing.assert_allclose(
            self.scene.get_frame(0).cameras["forward"].pose.rotation,
            [
                [4.16558418e-02, -6.63542562e-02, 9.96926228e-01],
                [-9.99126540e-01, 5.37968093e-04, 4.17835868e-02],
                [-3.30883332e-03, -9.97795983e-01, -6.62738888e-02],
            ],
        )

    def test_camera_intrinsic(self):
        np.testing.assert_allclose(
            self.scene.get_frame(0).cameras["forward"].K,
            [
                [338.56098087, 0.0, 638.95930485],
                [0.0, 335.39315139, 397.69312119],
                [0.0, 0.0, 1.0],
            ],
        )
        np.testing.assert_allclose(
            self.scene.get_frame(0).cameras["forward"].D,
            [
                0.08577354267758332,
                -0.009275696593772212,
                0.0,
                0.0,
                0.002280136164042399,
                -0.0008592168095191392,
            ],
        )


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



import pandas as pd
import ujson
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button
import numpy as np
from io import BytesIO
from pyquaternion import Quaternion
from typing import Any
from scipy.spatial.transform import Rotation as R
import open3d as o3d


from .camera import LidarCamera
from .image import LidarImage
from .connectors import Importer
from .transform import Transform
from .helper import (
    s3_smart_upload,
    format_lidar_point,
    format_point,
    format_quaternion,
    scale_file_upload,
)


class LidarFrame:
    """Frame object represents the point cloud, image and calibration data contained in a single frame

    Frame properties:
      - id: Frame id, used to identify the frame
      - cameras: List of LidarCamera
      - images: List of LidarImage
      - points: Pointcloud for this frame
      - radar_points: Radar points for this frame
      - colors: Colors for each point on the pointcloud for this frame
      - transform: Pose/ transform of this frame
    """

    def __init__(self, frame_id, cameras):
        self.id = frame_id
        self.cameras: pd.Series[Any, LidarCamera] = cameras
        self.images: pd.Series[Any, LidarImage] = pd.Series(dtype=object)
        self.points: np.ndarray = np.zeros((0, 5), dtype=float)
        self.radar_points: np.ndarray = np.zeros((0, 3), dtype=float)
        self.colors: np.ndarray = np.zeros((0, 3), dtype=float)
        self.transform = Transform()

    def get_image(self, camera_id) -> LidarImage:
        """Get image by camera_id or create one if it does not exist

        :param camera_id: Camera id
        :type camera_id: str, int
        :returns: LidarImage object
        :rtype: LidarImage

        """
        assert camera_id in self.cameras, "Camera not found"
        if camera_id not in self.images:
            if isinstance(camera_id, int):
                self.images.index = self.images.index.astype(int)
            self.images[camera_id] = LidarImage(camera=self.cameras[camera_id])
        return self.images[camera_id]

    def add_points_from_connector(
        self, connector: Importer, transform: Transform = None, intensity=1, sensor_id=0
    ):
        """Use Importer output to add points to the frame

        :param connector: Importer used to load the points
        :type connector: Importer
        :param transform: Transform that should be applied to the points
        :type transform: Transform
        :param intensity: If the points list does not include intensity, this value will be used as intensity for all the points (default ``1``)
        :type intensity: int
        :param sensor_id: Sensor id, used in case that you have more than one lidar sensor. (Default ``0``)
        :type sensor_id: int

        """
        self.add_points(connector.data, transform, intensity, sensor_id)

    def add_radar_points(self, points: np.array):
        """Add radar points to the frame, structure:

        .. highlight:: python
        .. code-block:: python

          radar_points = np.array([
            [
            [0.30694541, 0.27853175, 0.51152715],    // position - x,y,z
            [0.80424087, 0.24164057, 0.45256181],    // direction - x,y,z
            [0.73596422]    // size
            ],
            ...
          ])

        :param points: List of radar points data
        :type points: np.array

        """
        assert np.array(points).shape[1] == 3, "Radar points length is not 3"
        self.radar_points = points

    def add_points(
        self, points: np.array, transform: Transform = None, intensity=1, sensor_id=0
    ):
        """Add points to the frame, structure: np.array with dimension 1 and shape (N,3) or (N,4) (N being the number of point in the frame)

        Points with intensity:

        .. highlight:: python
        .. code-block:: python

          points = np.array([
            [0.30694541, 0.27853175, 0.51152715, 0.4],
            [0.80424087, 0.24164057, 0.45256181, 1],
            ...
          ])

        Points without intensity:

        .. highlight:: python
        .. code-block:: python

          points = np.array([
            [0.30694541, 0.27853175, 0.51152715],
            [0.80424087, 0.24164057, 0.45256181],
            ...
          ])

        :param points: List of points
        :type points: np.array
        :param transform: Transform that should be applied to the points
        :type transform: Transform
        :param intensity: If the points list doesn't include intensity, this value will be used as intensity for all the points (default ``1``)
        :type intensity: int
        :param sensor_id: Sensor id, used in case that you have more than one lidar sensor. (Default ``0``)
        :type sensor_id: int

        """
        if points.ndim == 1:
            points = np.array([points])
        if points.shape[1] == 3:
            points = np.hstack([points, np.ones((points.shape[0], 1)) * intensity])
        if transform is not None:
            points = transform.apply(points)

        points = np.hstack([points, np.ones((points.shape[0], 1)) * sensor_id])

        self.points = np.vstack([self.points, points])

    def add_colors(self, colors: np.ndarray):
        """Add colors to the pointcloud. This list should follow the same order as the point list.

        Each color should be in RGB with values between 0 and 255.

        .. highlight:: python
        .. code-block:: python

          colors = np.array([
            [10, 200, 230],
            [0, 0, 255],
            ...
          ])


        :param colors: List of colors
        :type colors: np.ndarray

        """
        self.colors = np.vstack([self.colors, colors])

    def add_debug_lines(self, intensity: int = 1, length: int = 5, device: int = 0):
        """Add debug lines.

        This will add a line starting from each camera position to the direction it is facing. This will use the camera position in this frame.

        :param intensity: Intensity of the points from the debugging line, default ``1``
        :type intensity: int
        :param length: Length of the line, default ``5`` points
        :type length: int
        :param device: Device id fror the points added, default ``0``
        :type device: int

        """
        x_line = np.array([np.array([length * k / 100, 0, 0]) for k in range(0, 100)])
        for camera in self.cameras:
            self.add_points(
                x_line, transform=camera.world_transform, intensity=intensity
            )

    def get_world_points(self):
        """Return the list of points with the frame transformation applied

        :returns: List of points in world coordinates
        :rtype: np.array

        """
        return np.hstack(
            [
                self.transform @ self.points[:, :3],
                self.points[:, 3:4],
                self.points[:, 4:5],
            ]
        )

    # leeaving this method as legacy / old code dependency
    def get_projected_image(self, camera_id, color_mode: str = "default", **kwargs):
        """Get camera_id image with projected points

        :param camera_id: Camera id/Name/Identifier
        :type camera_id: str, int
        :param color_mode:  Color mode, default ``default``, modes are: 'depth', 'intensity' and 'default'
        :type color_mode: str
        :returns: Image with the points projected
        :rtype: PIL.Image

        """
        return self.cameras[camera_id].get_projected_image(
            self.get_image(camera_id), self.points, self.transform, color_mode, **kwargs
        )

    def manual_calibration(
        self, camera_id, intrinsics_ratio: int = 1000, extrinsics_ratio: int = 10
    ):
        """Open a window with the camera with the points projected over it. The window also display dials to change the camera intrinsic and extrinsic values. The new values for the camera calibration will be display as matrices on the terminal.

        :param camera_id: Camera id/Name/Identifier
        :type camera_id: str, int
        :param intrinsics_ratio: Range of possible values for the intrinsic, center value will be the current one.
        :type intrinsics_ratio: int
        :param extrinsics_ratio:  Range of possible values for the extrinsic, center value will be the current one.
        :type extrinsics_ratio: int

        """
        fig = plt.figure(constrained_layout=True)
        intrinsics = ["fx", "fy", "cx", "cy"]
        extrinsics_position = ["x", "y", "z"]
        extrinsics_heading = ["qw", "qx", "qy", "qz"]
        heights = np.concatenate(
            (np.array([7]), np.array([0.2 for x in range(0, 7)])), axis=0
        )
        gs = fig.add_gridspec(
            ncols=3,
            nrows=1 + len(extrinsics_heading) + len(extrinsics_position),
            height_ratios=heights,
        )
        imgObj = fig.add_subplot(gs[0, :])
        imgObj.imshow(
            self.cameras[camera_id].get_projected_image(
                self.get_image(camera_id), self.points, self.transform, "depth", 1
            )
        )

        for index, key in enumerate(intrinsics):
            globals()[f"ax{key}"] = fig.add_subplot(gs[index + 1, 0])
            value = getattr(self.cameras[camera_id], key)
            globals()[key] = Slider(
                globals()[f"ax{key}"],
                f"{key}",
                value - intrinsics_ratio,
                value + intrinsics_ratio,
                valinit=value,
            )

        for index, key in enumerate(extrinsics_position):
            globals()[f"ax{key}"] = fig.add_subplot(gs[index + 1, 1])
            value = getattr(self.cameras[camera_id], "position")
            globals()[key] = Slider(
                globals()[f"ax{key}"],
                f"{key}",
                value[index] - extrinsics_ratio,
                value[index] + extrinsics_ratio,
                valinit=value[index],
            )

        for index, key in enumerate(extrinsics_heading):
            globals()[f"ax{key}"] = fig.add_subplot(
                gs[index + len(extrinsics_position) + 1, 1]
            )
            value = getattr(self.cameras[camera_id], "rotation")
            value = Quaternion(R.from_matrix(value).as_quat())
            globals()[key] = Slider(
                globals()[f"ax{key}"],
                f"{key}",
                value[index] - extrinsics_ratio,
                value[index] + extrinsics_ratio,
                valinit=value[index],
            )

        def update(val):
            self.cameras[camera_id].calibrate(
                K=np.array([[fx.val, 0, cx.val], [0, fy.val, cy.val], [0, 0, 1]]),
                pose=Transform.from_Rt(
                    Quaternion(qw.val, qx.val, qy.val, qz.val).rotation_matrix,
                    [x.val, y.val, z.val],
                ),
            )
            np.set_printoptions(suppress=True)
            print(f"New intrinsics for Camera {camera_id}")
            print(self.cameras[camera_id].K)
            print(f"New position for Camera {camera_id}")
            print(self.cameras[camera_id].position)
            print(f"New heading for Camera {camera_id}")
            print(
                dict(
                    zip(
                        ["w", "x", "y", "z"],
                        Transform(self.cameras[camera_id].rotation).quaternion,
                    )
                )
            )

            imgObj.imshow(
                self.cameras[camera_id].get_projected_image(
                    self.get_image(camera_id), self.points, self.transform, "depth", 1
                )
            )
            fig.canvas.draw_idle()

        axApply = fig.add_subplot(gs[1, 2])
        bApply = Button(axApply, "Apply changes")
        bApply.on_clicked(update)
        plt.show()

    def get_filename(self) -> str:
        """Get frame json file name

        :returns: Json file name
        :rtype: str

        """
        return "frame-%s.json" % self.id

    def apply_transform(self, T: Transform):
        """Apply the frame transformation. This will be used to define the device position and applied to cameras and points.

        :param T: Transform for this frame
        :type T: Transform

        """
        self.transform = Transform(T) @ self.transform

    def filter_points(self, min_intensity=None, min_intensity_percentile=None):
        """Filter points based on their intensity

        :param min_intensity: Minimun intensity allowed
        :type min_intensity: int
        :param min_intensity_percentile: Minimun percentile allowed (use np.percentile)
        :type min_intensity_percentile: int

        """
        if min_intensity is not None:
            self.points = self.points[self.points[:, 3] >= min_intensity]
        if min_intensity_percentile is not None:
            self.points = self.points[
                self.points[:, 3]
                >= np.percentile(self.points[:, 3], min_intensity_percentile)
            ]

    def to_json(
        self, base_url: str = "", s3_upload: bool = True, project_name: str = ""
    ):
        """Return the frame data in json format following Scale data format: https://private-docs.scale.com/?python#sensor-fusion-lidar-annotatio.

        This will return the final data from the frame, this means cameras and points will be in world coordinates.

        :param base_url: This url will concatenated with the image name, e.g.: `'%s/image-%s-%s.jpg' % (base_url, camera.id, frame.id)`
        :type base_url: str
        :returns: Frame object as a JSON formatted stream
        :rtype: str

        """

        def format_image(camera, s3_upload: bool, project_name: str):
            image = self.images[camera.id]
            if camera.world_poses:
                wct = camera.world_poses[self.id]
            else:
                wct = (image.transform or self.transform) @ camera.pose

            D = camera.D

            image_url = "%s/image-%s-%s.jpg" % (base_url, camera.id, self.id)
            if not s3_upload:
                image_url = self.images[camera.id].scale_file_upload(project_name)

            result = dict(
                position=format_point(wct.position),
                heading=format_quaternion(wct.quaternion),
                image_url=image_url,
                camera_model=camera.model,
                fx=camera.fx,
                fy=camera.fy,
                cx=camera.cx,
                cy=camera.cy,
                skew=camera.skew,
                k1=float(D[0]),
                k2=float(D[1]),
                p1=float(D[2]),
                p2=float(D[3]),
                k3=float(D[4]),
                k4=float(D[5]),
                k5=float(D[6]) if len(D) >= 7 else 0,
                k6=float(D[7]) if len(D) >= 8 else 0,
                scale_factor=camera.scale_factor,
            )
            if image.metadata:
                result["metadata"] = image.metadata
            if image.timestamp:
                result["timestamp"] = image.timestamp
            return result

        images_json = (
            self.cameras[self.images.index]
            .apply(format_image, args=(s3_upload, project_name))
            .to_json(orient="records")
        )
        points_json = pd.DataFrame(
            self.get_world_points(), columns=["x", "y", "z", "i", "d"]
        ).to_json(double_precision=4, orient="records", date_format=None)

        colors_json = pd.Series(self.colors.reshape(-1).astype(np.uint32)).to_json(
            orient="values", date_format=None
        )

        radar_points_json = ujson.dumps(
            list(
                np.array(
                    [
                        {
                            "position": format_point(row[0]),
                            "direction": format_point(row[1]),
                            "size": row[2][0],
                        }
                        for row in self.radar_points
                    ]
                )
            )
        )
        frame_object = {
            "images": "__IMAGES__",
            "points": "__POINTS__",
            "device_position": format_point(self.transform.position),
            "device_heading": format_quaternion(self.transform.quaternion),
        }

        if len(self.colors) > 0:
            frame_object["point_colors"] = "__COLORS__"

        if len(self.radar_points) > 0:
            frame_object["radar_points"] = "__RADAR_POINTS__"

        out = ujson.dumps(frame_object)
        out = out.replace('"__IMAGES__"', images_json)
        out = out.replace('"__POINTS__"', points_json)
        out = out.replace('"__COLORS__"', colors_json)
        out = out.replace('"__RADAR_POINTS__"', radar_points_json)

        return out

    def save(self, path: str, base_url: str = ""):
        """Save frame object in a json file

        :param path: Path in which the frame data should be saved
        :type path: str
        :param base_url: This url will concatenated with the image name, e.g.: `'%s/image-%s-%s.jpg' % (base_url, camera.id, frame.id)`
        :type base_url: str

        """
        # Save frame
        with open(os.path.join(path, "frame-%s.json" % self.id), "w") as file:
            file.write(self.to_json(base_url))

        # Save images
        for camera_id, image in self.images.items():
            image.save(os.path.join(path, "image-%s-%s.jpg" % (camera_id, self.id)))

    def s3_upload(self, bucket: str, path: str):
        """Save frame in S3

        :param bucket: S3 Bucket name
        :type bucket: str
        :param path: Path to store data
        :type key: str

        """
        # print(f'Uploading frame {self.id}...')
        base_url = f"s3://{bucket}/{path}"

        # Upload frame json file
        s3_smart_upload(
            fileobj=BytesIO(bytes(self.to_json(base_url), encoding="utf-8")),
            bucket=bucket,
            key=f"{path}/frame-{self.id}.json",
            content_type="application/json",
        )

        # Upload images
        for camera_id, image in self.images.items():
            image.s3_upload(bucket, f"{path}/image-{camera_id}-{self.id}.jpg")

    def scale_file_upload(self, project_name: str):
        """Save frame in Scale File

        :param project_name: File project name
        :type bucket: str

        """

        fileobj = BytesIO(
            bytes(
                self.to_json(s3_upload=False, project_name=project_name),
                encoding="utf-8",
            )
        )
        return scale_file_upload(fileobj, project_name)

    def downsample_frame(self, voxel_size_mm=250):
        pcd = o3d.geometry.PointCloud()
        normals = np.zeros((self.points.shape[0], 3), dtype=float)
        normals[:, 0] = self.points[:, 3]

        pcd.points = o3d.utility.Vector3dVector(self.points[:, :3])
        pcd.normals = o3d.utility.Vector3dVector(normals)
        pcd.colors = o3d.utility.Vector3dVector(self.colors)

        pcd = pcd.voxel_down_sample(voxel_size_mm / 1000)

        positions, intensities = np.asarray(pcd.points), np.asarray(pcd.normals)[:, 0]
        points = np.hstack([positions, intensities.reshape(intensities.shape[0], 1)])

        self.points = np.hstack([points, np.zeros((points.shape[0], 1), dtype=float)])
        self.colors = np.asarray(pcd.colors)

    def __repr__(self):
        return "Frame({0}) {1}".format(self.id, self.transform)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scale_lidar_io",
    version="0.1.200",
    author="Scale AI",
    author_email="rodrigo.belfiore@scale.com, ivan.roumec@scale.com",
    description="Lidar data conversion helpers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["scale_lidar_io"],
    install_requires=[
        "scaleapi>=2.0.3",
        "pytest",
        "pymongo",
        "open3d",
        "cryptography",
        "requests",
        "awscli",
        "smart_open",
        "boto3",
        "dataclasses",
        "laspy",
        "matplotlib",
        "numpy",
        "opencv-python",
        "open3d",
        "pandas",
        "Pillow",
        "PyYAML",
        "transforms3d",
        "ujson",
        "pyquaternion",
        "pyntcloud",
        "urllib"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
